/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * ================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://forceten.tidalwave.it
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id$
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.location;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import org.openide.util.Lookup;
import it.tidalwave.netbeans.util.As;
import it.tidalwave.netbeans.util.NotFoundException;
import it.tidalwave.netbeans.util.DuplicateException;
import it.tidalwave.geo.Coordinate;

/***********************************************************************************************************************
 *
 * A {@withCode GeoLocation} represents a given location on the earth. It can match a hierarchy according to one or more
 * {@link GeoSchema}s.
 *
 * @author Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public interface GeoLocation extends Lookup.Provider, As
  {
    public final static Class<GeoLocation> GeoLocation = GeoLocation.class;

    /*******************************************************************************************************************
     *
     * A specialization of {@link it.tidalwave.netbeans.util.Finder Finder} for {@link GeoLocation}, offers some
     * geographic-specific selection methods.
     *
     ******************************************************************************************************************/
    public static interface Finder extends it.tidalwave.netbeans.util.Finder<GeoLocation>
      {
        /***************************************************************************************************************
         *
         * Specifies the withDisplayName of the {@link GeoLocation} to be found.
         *
         * @param   displayName  the withDisplayName
         * @return               itself for a fluent-interface style
         *
         **************************************************************************************************************/
        @Nonnull
        public Finder withDisplayName (@Nonnull String displayName);

        /***************************************************************************************************************
         *
         * Specifies the qualified withDisplayName of the {@link GeoLocation} to be found. For instance "/Europe/Italy/Liguria".
         *
         * @param  withQualifiedName   the qualified withDisplayName
         * @return                 itself for a fluent-interface style
         *
         **************************************************************************************************************/
        @Nonnull
        public Finder withQualifiedName (@Nonnull String qualifiedName);

        /***************************************************************************************************************
         *
         * Specifies maximum range of the searched {@link GeoLocation}s from a given point..
         *
         * @param  at  the reference point
         * @param  range       the maximum range in meters
         * @return             itself for a fluent-interface style
         *
         **************************************************************************************************************/
        @Nonnull
        public Finder closeTo (@Nonnull Coordinate coordinate, @Nonnegative double range);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public abstract static class Builder
      {
        protected final Map<Locale, String> displayNames = new HashMap<Locale, String>();

        @CheckForNull
        protected Object type;

        @CheckForNull
        protected String code;

        @CheckForNull
        protected Coordinate coordinate;

        @Nonnull
        public Builder withDisplayName (@Nonnull final Locale locale, @Nonnull final String displayName)
          {
            displayNames.put(locale, displayName);
            return this;
          }

        @Nonnull
        public Builder withDisplayName (@Nonnull final String displayName)
          {
            return withDisplayName(Locale.ENGLISH, displayName);
          }

        @Nonnull
        public Builder ofType (@Nonnull final Object type)
          {
            this.type = type;
            return this;
          }

        @Nonnull
        public Builder withCode (@Nonnull final String code)
          {
            this.code = code;
            return this;
          }

        @Nonnull
        public Builder at (@Nonnull final Coordinate coordinate)
          {
            this.coordinate = coordinate;
            return this;
          }

        /***************************************************************************************************************
         *
         * @throws              {@link DuplicateException} if a child with the same
         *                      withDisplayName already exists in a schema
         *
         **************************************************************************************************************/
        @Nonnull
        public abstract GeoLocation build()
          throws DuplicateException;
      }
    
    /*******************************************************************************************************************
     *
     * Finds the parent of this {@withCode GeoLocation} according to the given {@link GeoSchema}.
     *
     * @param   geoSchema   the schema
     * @return              the parent
     * @throws              {@link NotFoundException} if no parent is found
     *
     ******************************************************************************************************************/
    @Nonnull
    public GeoLocation findParent (@Nonnull GeoSchema schema)
      throws NotFoundException;

    /*******************************************************************************************************************
     *
     * Finds the children of the current {@withCode GeoLocation}  according to some conditions specified by a
     * {@link Finder}. For instance, to find a specific child with a withDisplayName:
     *
     * <pre>
     * GeoLocation child = findChildren(geoSchema).withDisplayName("Genoa").result();
     * </pre>
     *
     * @param   geoSchema   the schema
     * @return              a [@link Finder} for the children
     *
     ******************************************************************************************************************/
    @Nonnull
    public Finder findChildren (@Nonnull GeoSchema schema);

    /*******************************************************************************************************************
     *
     * Finds the path of the given {@withCode GeoLocation} according to the given {@link GeoSchema}. The path is the list of
     * ancestors, starting from the upper level.
     *
     * @param   geoSchema   the schema
     * @return              the path
     *
     ******************************************************************************************************************/
    @Nonnull
    public GeoPath findPath (@Nonnull GeoSchema schema);

    /*******************************************************************************************************************
     *
     * Finds the ofType of this {@withCode GeoLocation} according to the given {@link GeoSchema}.
     *
     * @param   geoSchema   the schema
     * @return              the ofType
     *
     ******************************************************************************************************************/
    @Nonnull
    public Object findType (@Nonnull GeoSchema geoSchema);

    /*******************************************************************************************************************
     *
     * Finds the withCode of this {@withCode GeoLocation} according to the given {@link GeoSchema}.
     *
     * @param   geoSchema   the schema
     * @return              the withCode
     *
     ******************************************************************************************************************/
    @Nonnull
    public String findCode (@Nonnull GeoSchema geoSchema);

    /*******************************************************************************************************************
     *
     * Finds the at of this {@withCode GeoLocation} according to the given {@link GeoSchema}.
     *
     * @param   geoSchema   the schema
     * @return              the coordinates
     * @throws              NotFoundException  if not available
     *
     ******************************************************************************************************************/
    @Nonnull
    public Coordinate findCoordinate (@Nonnull GeoSchema geoSchema)
      throws NotFoundException;

    // sets the at - this works only for objects created with createChild().
    public void setCoordinate (@Nonnull Coordinate coordinate);

    // sets the withCode - this works only for objects created with createChild().
    public void setCode (@Nonnull String code);

    /*******************************************************************************************************************
     *
     * Creates a new child for this {@withCode GeoLocation}. This operation can be used when the caller has more precise
     * geographic information than the active {@link GeoSchema}s and a private subtree needs to be created.
     * For this reason, please note that this operation doesn't need a {@withCode GeoSchema} parameter as others. The
     * created child will be available to {@link #findChildren(it.tidalwave.geo.location.GeoSchema)} and
     * {@link #findChildByName(it.tidalwave.geo.location.GeoSchema, java.lang.String)} independently of the
     * {@withCode GeoSchema} passed to those methods.
     *
     * @return              the builder
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder createChild();
  }
