/***********************************************************************************************************************
 *
 * forceTen - open source geography
 * ================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://forceten.tidalwave.it
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id$
 *
 **********************************************************************************************************************/
package it.tidalwave.geo.location;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.io.Serializable;
import org.openide.util.Lookup;
import it.tidalwave.netbeans.util.As;
import it.tidalwave.netbeans.util.NotFoundException;
import it.tidalwave.netbeans.capabilitiesprovider.LookupFactory;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.Sector;

/***********************************************************************************************************************
 *
 * This class models an immutable set of unordered {@link GeoLocation}s.
 * 
 * @author Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Immutable
public final class GeoLocationSet implements Serializable, Lookup.Provider, As, Iterable<GeoLocation>
  {
    private static final long serialVersionUID = 69392340982359L;

    @CheckForNull
    private Lookup lookup;

    private final Collection<GeoLocation> locations;

    private final Sector sector;

    @CheckForNull
    private GeoSchema geoSchema;

    /*******************************************************************************************************************
     *
     * Creates an empty set of {@link GeoLocation}s.
     *
     ******************************************************************************************************************/
    public GeoLocationSet()
      {
        locations = Collections.<GeoLocation>emptyList();
        sector = new Sector();
      }
    
    /*******************************************************************************************************************
     *
     * Creates a set of {@link GeoLocation}s with the given contents.
     *
     ******************************************************************************************************************/
    public GeoLocationSet (final @Nonnull GeoSchema geoSchema,
                           final @Nonnull GeoLocation ... locations)
      {
        this(geoSchema, Arrays.asList(locations));
      }

    /*******************************************************************************************************************
     *
     * Creates a set of {@link GeoLocation}s with the given contents.
     *
     ******************************************************************************************************************/
    public GeoLocationSet (final @Nonnull GeoSchema geoSchema,
                           final @Nonnull Collection<GeoLocation> locations)
      {
        this.geoSchema = geoSchema;
        this.locations = Collections.unmodifiableCollection(locations);
        final List<Coordinate> coordinates = new ArrayList<Coordinate>();

        for (final GeoLocation location : locations)
          {
            try
              {
                coordinates.add(location.findCoordinate(geoSchema));
              }
            catch (NotFoundException e)
              {
                // FIXME
              }
          }

        sector = new Sector(coordinates);
      }

    /*******************************************************************************************************************
     *
     * Creates a new set of {@link GeoLocation}s adding the given {@link GeoLocation}.
     *
     * @param   location  the {@link GeoLocation} to add
     * @return  a new set
     *
     ******************************************************************************************************************/
    @Nonnull
    public GeoLocationSet with (final @Nonnull GeoLocation location)
      {
        final List<GeoLocation> all = new ArrayList<GeoLocation>(this.locations);
        all.add(location);
        return new GeoLocationSet(geoSchema, all);
      }

    /*******************************************************************************************************************
     *
     * Creates a new set of {@link GeoLocation}s adding the given {@link GeoLocation}s.
     *
     * @param   {@link GeoLocation}  the {@link GeoLocation} to add
     * @return  a new set
     *
     ******************************************************************************************************************/
    @Nonnull
    public GeoLocationSet with (final @Nonnull Collection<GeoLocation> location)
      {
        final List<GeoLocation> all = new ArrayList<GeoLocation>(this.locations);
        all.addAll(location);
        return new GeoLocationSet(geoSchema, all);
      }

    /*******************************************************************************************************************
     *
     * Creates a new set of {@link GeoLocation}s adding the given {@link GeoLocation}s.
     *
     * @param   {@link GeoLocation}  the {@link GeoLocation} to add
     * @return  a new set
     *
     ******************************************************************************************************************/
    @Nonnull
    public GeoLocationSet with (final @Nonnull GeoLocation ... locations)
      {
        return with(Arrays.asList(locations));
      }

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public GeoSchema getGeoSchema()
      {
        return geoSchema;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Iterator<GeoLocation> iterator()
      {
        return Collections.unmodifiableCollection(locations).iterator();
      }

    /*******************************************************************************************************************
     *
     * Returns the size of the set.
     *
     * @return  the size of the set
     *
     ******************************************************************************************************************/
    @Nonnegative
    public int getSize()
      {
        return locations.size();
      }

    /*******************************************************************************************************************
     *
     * Returns the smaller {@link Sector} containing all the {@link GeoLocation}s.
     *
     * @return  the {@code Sector}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Sector getSector()
      {
        return sector;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public <T> T as (final @Nonnull Class<T> clazz)
      {
        return as(clazz, As.Defaults.throwAsException(clazz));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public <T> T as (final @Nonnull Class<T> clazz,
                     final @Nonnull NotFoundBehaviour<T> notFoundBehaviour)
      {
        final T object = getLookup().lookup(clazz);
        return (object != null) ? object : notFoundBehaviour.run(null);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public synchronized Lookup getLookup()
      {
        if (lookup == null)
          {
            // FIXME: use a LazyLookup for GeoContainerImpl?
//            lookup = LookupFactory.createLookup(this, Lookups.fixed(new GeoContainerImpl(this)));
            lookup = LookupFactory.createLookup(this);
          }

        return lookup;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean equals (final @CheckForNull Object object)
      {
        if (object == null)
          {
            return false;
          }

        if (getClass() != object.getClass())
          {
            return false;
          }

        final GeoLocationSet other = (GeoLocationSet)object;

        return this.locations.equals(other.locations) && this.sector.equals(other.sector);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public int hashCode()
      {
        int hash = 5;
        hash = 97 * hash + locations.hashCode();
        hash = 97 * hash + sector.hashCode();
        return hash;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String toString()
      {
        return sector.isVoid() ? "GeoLocationSet[empty]"
                               : String.format("GeoLocationSet[%s / %s]", locations, sector);
      }
  }
